//**************************************************************************//
//**                                                                      **//
//**                            WinNES                                    **//
//**       Nintendo Entertainment System Emulator for Windows             **//
//**                                                                      **//
//**   Copyright (C) Dott. Cosimo De Michele 1998                         **//
//**         You are not allowed to distribuite this software             **//
//**         commercially. Please, notify me, if you make any             **//
//**         changes to this file.                                        **//
//**                                                                      **//
//**  Nintendo is a trademark of Nintendo inc.                            **//
//**************************************************************************//

unit NesForm;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Menus, ComCtrls, StdCtrls, ExtCtrls, DXClass, DXDraws, DIB, DXSprite,
  DXInput, DXSounds, VersionInfo, MF6502, MMSystem;

type
  TPalType = array[0..63, 1..3] of byte;

  TMidiMsg = Record
    Status,
    Data1,
    Data2,
    NoData: Byte;
  end;

  TPsrNotes = (Piano,
               FlangePiano,
               HonkyTonkPiano,
               ElectricPiano1,
               ElectricPiano2,
               ElectricPinao3,
               Harpsichord1,
               Harpsichord2,
               Clavi,
               Celesta,
               PopeOrgan1,
               PipeOrgan2,
               ElectricOrgan1,
               ElectricOrgan2,
               ElectricOrgan3,
               ElectricOrgan4,
               Accordian1,
               Accordian2,
               ElectricGuitar1,
               ElectricGuitar2,
               ElectricGuitar3,
               TremoloGuitar,
               Electric12StringGuitar,
               DistortionGuitar,
               JazzGuitar,
               JazzGuitarOctave,
               MuteGuitar,
               MuteGuitarEcho,
               SteelGuitar,
               FolkGuitar,
               TwelveStringGuitar,
               GutGuitar,
               Violin1,
               Violin2,
               Cello,
               Strings1,
               Stings2,
               OrchestraHit,
               Harp,
               Bango,
               Vibraphone,
               Marima,
               SteelDrum,
               Trumpet,
               MuteTrumpet1,
               MuteTrumpet2,
               MuteTrumpet3,
               Trombone,
               FlugelHone,
               Norn,
               Tuba,
               BrassEnsemble1,
               BrassEnsemble2,
               Piccolo,
               Flute,
               Clarinet,
               BassClarinet,
               Oboe,
               EnglishHorn,
               Bassoon,
               SopranoSax,
               AltoSax,
               TenorSax,
               BraitoneSax,
               Ocarina,
               PanFlute,
               Recorder,
               Harmonica,
               SambaWhistle,
               SaxEnsemble1,
               SaxEnsemble2,
               WoodWindEnsemble,
               Chorus,
               SynthLead,
               SynthBrass,
               SynthStrings,
               SynthTom,
               Fantasy1,
               Fantasy2,
               Fantasy3,
               BellStrings,
               SeqPad,
               ElectricBass1,
               ElectricBass2,
               FretlesBass,
               MuteBass,
               MuteBassEcho,
               SlapBass,
               WoodBass1,
               WoodBass2,
               SynthBass1,
               SynthBass2,
               SynthBass3,
               BowedBass,
               Scratch,
               KickAndSnare,
               Tom,
               LatinPercs,
               PercussionsWithGate,
               Percussions);

const
  NES_SCREEN_WIDTH  = 256;
  NES_SCREEN_HEIGHT = 240;
  MINLINE           = 0;
  MAXLINE           = 238;
  SCANLINEPERFRAME  = 244;
  SNDPERIOD         = 244;
  NO_DATA           = $FF;

  ROM_MIRROR  = 1;
  ROM_WRAM    = 2;
  ROM_TRAINER = 4;
  ROM_FOUR    = 8;

  PPU_InVBlank    = $80;
  PPU_SpHit       = $40;
  PPU_SpMax       = $20;
  PPU_WriteIgnore = $10;

  JOY_A      = $01;
  JOY_B      = $02;
  JOY_SELECT = $04;
  JOY_START  = $08;
  JOY_UP     = $10;
  JOY_DOWN   = $20;
  JOY_LEFT   = $40;
  JOY_RIGHT  = $80;

  V_FLIP     = $80;
  H_FLIP     = $40;
  SP_BACK    = $20;
  PSG_BASE   = 111860;

  BLACK_COLOR = 63;  { 63 is the entry of black in palette }
  MRUCount   = 5;
  RegKey     = 'Software\Karakiri\WinNES';
  midiNotePlay = $90; { 144 }
  midiChangeInstrument = $C0; { 192 }
  SndVolume = 64;

  Pal : TPalType = {Alex Krasivsky's Palette}
    (
    ($68,$68,$68),($00,$40,$80),($00,$00,$80),($40,$00,$80),
    ($80,$00,$80),($80,$00,$40),($80,$00,$00),($55,$00,$00),
    ($40,$40,$00),($33,$50,$00),($00,$50,$00),($00,$50,$00),
    ($00,$40,$40),($00,$00,$00),($00,$00,$00),($00,$00,$00),
    ($98,$98,$98),($00,$80,$C0),($40,$40,$C0),($80,$00,$C0),
    ($C0,$00,$C0),($C0,$00,$80),($C0,$20,$20),($C0,$40,$00),
    ($80,$80,$00),($55,$80,$00),($00,$80,$00),($33,$80,$33),
    ($00,$80,$80),($00,$00,$00),($00,$00,$00),($00,$00,$00),
    ($D0,$D0,$D0),($40,$C0,$FF),($80,$80,$FF),($C0,$40,$FF),
    ($FF,$00,$FF),($FF,$40,$C0),($FF,$50,$50),($FF,$80,$40),
    ($C0,$C0,$00),($80,$C0,$00),($00,$C0,$00),($55,$C0,$55),
    ($00,$C0,$C0),($33,$33,$33),($00,$00,$00),($00,$00,$00),
    ($FF,$FF,$FF),($80,$FF,$FF),($C0,$C0,$FF),($FF,$80,$FF),
    ($FF,$40,$FF),($FF,$80,$FF),($FF,$80,$80),($FF,$C0,$80),
    ($FF,$FF,$40),($C0,$FF,$40),($40,$FF,$40),($AA,$FF,$AA),
    ($40,$FF,$FF),($99,$99,$99),($00,$00,$00),($00,$00,$00)
    );

type
  TRomHeader = record
    RomType : array[1..3] of char;
    Ctrl,
    PRGROM,
    CHRRAM,
    ROMctrl1,
    ROMctrl2 : char;
    Expansion : array[0..7] of char;
  end;

  TfrmNES = class(TDXForm)
    MainMenu: TMainMenu;
    mnuFile: TMenuItem;
    mnuSetting: TMenuItem;
    mnuLoadROM: TMenuItem;
    mnuFreeROM: TMenuItem;
    mnuRom: TMenuItem;
    N2: TMenuItem;
    mnuExit: TMenuItem;
    Help1: TMenuItem;
    mnuAbout: TMenuItem;
    StatusBar: TStatusBar;
    OpenDialog: TOpenDialog;
    mnuInput: TMenuItem;
    N3: TMenuItem;
    Terminal: TMemo;
    mnuCPU: TMenuItem;
    mnuReset: TMenuItem;
    N4: TMenuItem;
    mnuPause: TMenuItem;
    mnuDebug: TMenuItem;
    DXDIB: TDXDIB;
    DXInput1: TDXInput;
    mnuAudio: TMenuItem;
    mnuNetwork: TMenuItem;
    DXInput2: TDXInput;
    VersionResource: TVersionResource;
    mnuSavesnapshot: TMenuItem;
    SaveDialog: TSaveDialog;
    MRUSeparator: TMenuItem;
    MRU1: TMenuItem;
    MRU2: TMenuItem;
    MRU3: TMenuItem;
    MRU4: TMenuItem;
    MRU5: TMenuItem;
    DXDraw: TDXDraw;
    DXImageList: TDXImageList;
    PopupMenu: TPopupMenu;
    pmuLoadRom: TMenuItem;
    pmuFreeROM: TMenuItem;
    N1: TMenuItem;
    pmuReset: TMenuItem;
    pmuPause: TMenuItem;
    procedure mnuLoadROMClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure mnuExitClick(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure mnuDebugClick(Sender: TObject);
    procedure mnuPauseClick(Sender: TObject);
    procedure mnuResetClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure mnuRomClick(Sender: TObject);
    procedure mnuAboutClick(Sender: TObject);
    procedure mnuFreeROMClick(Sender: TObject);
    procedure mnuInputClick(Sender: TObject);
    procedure mnuSavesnapshotClick(Sender: TObject);
    procedure MRUClick(Sender: TObject);
    procedure mnuFileClick(Sender: TObject);
    procedure mnuCPUClick(Sender: TObject);
    procedure mnuFullScreenClick(Sender: TObject);
  private
    CPUPeriod,
    UPeriod,
    UCount,
    SndCount,

    PrevLine,
    HitY : integer;

    ScreenWidth,
    ScreenHeight,
    CPUinstr,
    FrameRate,
    ScanLine,
    LastTimerTick,
    CurrentFrame : integer;
    { Sound }
    SoundOn : boolean;
    hMidi   : PhMidiOut;
    hOut    : THandle;
    Msg     : TMidiMsg;

    { Cartridge setting }
    RomFilename : string;
    RomHeader   : TRomHeader;
    { Joypad 1 & 2 }
    Joypad1,
    Joypad2,
    Joypad1Read,
    Joypad2Read  : byte;
    { PPU register }
    PPU_PalChange,
    PPUSCROLL_Change,
    PPUADDR_FirstRead,
    PPUSCROLL_Toggle : boolean;
    PPUSCROLL_X1,
    PPUSCROLL_Y1,
    PPUSCROLL_X2,
    PPUSCROLL_Y2 : byte;
    MRUList : TStringList;
    { PSG register }
    SndFreq : array[0..3] of byte;
    SndFreqVal : array[0..3] of word;
    SndOff : array[0..3] of boolean;
    SndLen : array[0..3] of integer;
    { }
    procedure RunMachine(Cartridge : string);
    function  LoadNES(filename : string): boolean;
    function  InitNES: boolean;
    procedure ResetNES;
    procedure TrashMachine;
    procedure IOWrite(A : word; V : byte);
    function  IORead(A : word): byte;
    procedure CheckKb;
    procedure RefreshLine (Y1, Y2 : integer);
    procedure RefreshSprite(Y1, Y2 : integer);
    procedure InitScreen;
    procedure DrawScreen;
    procedure TrashScreen;
    procedure ReadConfig;
    procedure WriteConfig;
    procedure SetFreq(ch, freq, vol : WORD);
    procedure SetVolume(ch, vol : integer);
    procedure PlayLogo;
  public
    cpu : M6502;
    NESisRunning : boolean;
    Page,
    VPage : array [0..8] of PChar;
    { Cartridge ROM }
    ROM,
    VROM  : PChar;
    VAddr : word;
    { PPU }
    PPU    : array[0..3] of byte;
    VRAM   : array[0..$4000] of byte;
    SRAM   : array[0..$0100] of byte;
    WRAM   : array[0..$1FFF] of byte;
    { PSG }
    PSG    : array [0..$17] of byte;

    ROMSize,
    VROMSize,
    ROMType,
    MMCType,
    Mirroring,
    MirrorXor  : integer;
    ViewFreq : boolean;
    { NES Keyboard }
    BUTTON1_UP,
    BUTTON1_DOWN,
    BUTTON1_LEFT,
    BUTTON1_RIGHT,
    BUTTON1_A,
    BUTTON1_B,
    BUTTON1_START,
    BUTTON1_SELECT,
    BUTTON2_UP,
    BUTTON2_DOWN,
    BUTTON2_LEFT,
    BUTTON2_RIGHT,
    BUTTON2_A,
    BUTTON2_B,
    BUTTON2_START,
    BUTTON2_SELECT,
    WIN_MODE  : integer;
    BUTTON1_ENABLED,
    BUTTON2_ENABLED : boolean;
    SCREEN_MODE : string;
    { Central Memory }
    RAM : array [0..$800] of byte;

    function MemoryReadOpcode(Addr : word) : byte;
    function MemoryReadByte(Addr : word) : byte;
    procedure MemoryWrite (Addr : word; value : byte);
    function cpuLoop : byte;
  end;

var
  frmNES: TfrmNES;
  Version : string;

implementation

uses Registry, RomPropertyForm, AboutForm, SettingForm,
  Mapper; //DebugForm

{$R *.DFM}

procedure AssignKey(var KeyAssignList: TKeyAssignList; State: TDXInputState;
  const Keys: array of Integer);
var
  i, i2: Integer;
  KeyAssign: PKeyAssign;
begin
  KeyAssign := @KeyAssignList[State];
  FillChar(KeyAssign^, SizeOf(TKeyAssign), 0);
  i2 := 0;
  for i:=LOW(Keys) to HIGH(Keys) do
  begin
    if i2<3 then
      KeyAssign^[i2] := Keys[i]
    else
      Exit;
    Inc(i2);
  end;
end; { AssignKey }

procedure TfrmNES.SetFreq(ch, freq, vol : WORD);
var
  B : byte;
  f : longint;
begin
  //if not SoundOn then
  //  Exit;

  if freq>6208 then
    B := 8
  else if freq>3104 then
    B := 7
  else if freq>1552 then
    B := 6
  else if freq>776 then
    B := 5
  else if freq>388 then
    B := 4
  else if freq>194 then
    B := 3
  else if freq>97 then
    B := 2
  else
    B := 1;

  f := (freq shl (20 - B)) div 49716;
  caption := 'ch : ' + inttostr(ch)+ ' freq : '+ inttostr(freq) + ' octava : ' + inttostr(B) + ' f : ' + inttostr(f and $FF) + ' len : ' + inttostr(sndlen[ch]);
  { setto lo strumento }
  Msg.Status := midiChangeInstrument; { $C0; }
  Msg.Data1 := ORD(SynthLead);
  Msg.Data2 := 0;
  Msg.NoData := 0;
  MidiOutShortMsg(hMidi^, LongInt(Msg));
  { setto la nota }
  Msg.Status := midiNotePlay  + ch;
  Msg.Data1 := f and $FF;
  Msg.Data2 := vol;
  Msg.NoData := 0;
  MidiOutShortMsg(hMidi^, LongInt(Msg));
end; { SetFreq }

procedure TfrmNES.SetVolume(ch, vol : integer);
begin
end; { SetVolume }

procedure TfrmNES.RunMachine(Cartridge : string);
begin
  mnuFreeROM.Click;
  if LoadNES(Cartridge) then
  begin
    if InitNES then
    begin
      mnuRom.enabled := true;
      mnuFreeROM.enabled := true;
      mnuSaveSnapshot.enabled := true;
      { reset the NES machine }
      ResetNes;
      Terminal.Lines.Add('6502 engine started.');
      { reset the sound }
      New(hMidi);
      SoundOn := (MidiOutOpen(hMidi, $FFFF, 0, 0, 0)=0);
      { run the NES machine }
      InitScreen;
      NESisRunning := true;
      Run6502(cpu);
    end;
  end;
end; { RunMachine }

procedure TfrmNES.mnuLoadROMClick(Sender: TObject);
begin
  if doFullScreen in DXDraw.Options then
    mnuFreeROM.Click;
  if OpenDialog.Execute then
  begin
    { update MRU list }
    if MRUList.indexOf(OpenDialog.Filename)=-1 then
    begin
      MRUList.Insert(0, OpenDialog.Filename);
      if MRUList.Count>MRUCount then
        MRUList.Delete(MRUCount);
    end;
    { run the NES }
    RunMachine(OpenDialog.Filename);
  end;
end; { mnuLoadROMClick }

procedure TfrmNES.ResetNES;
begin
  UPeriod := 2;
  UCount  := 0;
  { Joypads }
  Joypad1 := 0;
  Joypad2 := 0;
  Joypad1Read := 0;
  Joypad2Read := 0;
  { PPU register }
  PPU[0] := 0;
  PPU[1] := 0;
  PPU[2] := 0;
  PPU[3] := 0;
  VAddr  := 0;
  { Screen }
  PPUADDR_FirstRead  := true;
  PPUSCROLL_Toggle := false;
  scanline := 0;
  Terminal.Lines.Add('Resetting 6502 engine ...');
  { Reset cpu }
  CPUinstr := 0;
  Reset6502(cpu);
end; { ResetNES }

procedure TfrmNES.TrashMachine;
var
  i : integer;
  BatteryName : string;
  BatteryFile : file;
begin
  if (ROMType and ROM_WRAM)<>0 then
  begin
    { Save backed memory }
    BatteryName := Copy(RomFilename, 1, Pos('.', RomFilename)) + 'sav';
    Terminal.Lines.Add('Saving backed memory file "' + ExtractFilename(BatteryName) + '".');
    AssignFile(BatteryFile, BatteryName);
    ReWrite(BatteryFile, 1);
    BlockWrite(BatteryFile, WRAM, SizeOf(WRAM), i);
    CloseFile(BatteryFile);
  end;
  { azzero memoria }
  FillChar(VRAM, SizeOf(VRAM), $00);
  FillChar(SRAM, SizeOf(SRAM), $00);
  FillChar(WRAM, SizeOf(WRAM), $00);
  FillChar(RAM, SizeOf(RAM), $00);
  for i:=Low(Page) to High(Page) do
    Page[i] := nil;
  for i:=Low(VPage) to High(VPage) do
    VPage[i] := nil;
  if Assigned(ROM) then
  begin
    FreeMem(ROM);
    ROM := nil;
  end;
  if Assigned(VROM) then
  begin
    FreeMem(VROM);
    VROM := nil;
  end;
  RomFilename := '';
  { sound }
  MidiOutClose(hMidi^);
  Dispose(hMidi);
  { ripristino lo schermo }
  TrashScreen;
  Terminal.Lines.Add('Cartridge discharged.');
end; { TrashMachine }

function TfrmNES.InitNES: boolean;
var
  j : integer;
  BatteryName : string;
  BatteryFile : file;
begin
  if (ROMType and ROM_WRAM)<>0 then
  begin
    Page[3] := @WRAM;
    { Loading of backed memory }
    Terminal.Lines.Add('A backed memory has been found.');
    BatteryName := Copy(RomFilename, 1, Pos('.', RomFilename)) + 'sav';
    if FileExists(BatteryName) then
    begin
      Terminal.Lines.Add('Loading backed memory file "' + ExtractFilename(BatteryName) + '".');
      AssignFile(BatteryFile, BatteryName);
      Reset(BatteryFile, 1);
      BlockRead(BatteryFile, WRAM, SizeOf(WRAM), j);
      CloseFile(BatteryFile);
    end;
  end;
  Page[0] := @RAM;
  Result := MMCInit(MMCType);
end; { InitNES }

function TfrmNES.LoadNES(filename : string) : boolean;
var
  F : file;
  i, j,
  PRGBytes,
  CHRBytes,
  CartSize : integer;
begin
  frmNES.Terminal.Lines.Add('');
  RomFilename := ExtractFilename(filename);
  Terminal.Lines.Add('Loading Cartridge : "' + RomFileName +'".');
  AssignFile(F, filename);
  Reset(F, 1);
  { Header }
  BlockRead(F, ROMHeader, SizeOf(ROMHeader), i);
  if (ROMHeader.RomType <> 'NES') then
  begin
   Terminal.Lines.Add('Unsupported file format.');
   Result := false;
   Exit;
  end;
  Terminal.Lines.Add('iNES file format found.');

  ROMSize  := Ord(ROMHeader.PRGROM);
  VROMSize := Ord(ROMHeader.CHRRAM);
  ROMType  := Ord(ROMHeader.ROMctrl1);
  MMCType  := (Ord(ROMHeader.ROMctrl1) shr 4);
  CartSize := ROMSize shl 4 + VROMSize * 8;

  Terminal.Lines.Add(format('[%d] of PRG Rom', [ROMSize]));
  Terminal.Lines.Add(format('[%d] of CHR Ram', [VROMSize]));
  Terminal.Lines.Add(format('Size of the cartridge : [%d Kb]', [CartSize]));

  if ROMHeader.Expansion[0]=#0 then
    MMCType := MMCType or (Ord(ROMHeader.ROMctrl2) and $F0);
  Mirroring := (Ord(ROMHeader.ROMctrl1) and ROM_MIRROR);
  MirrorXor := $400 shl Mirroring; { Horizontal : $400; Vertical : $800)
  { dimensionamento aree di memoria }
  PRGBytes := $4000 * ROMSize;
  CHRBytes := $2000 * VROMSize;
  GetMem(ROM, PRGBytes + 1);
  GetMem(VROM, CHRBytes + 1);
  { Caricamento della ROM }
  BlockRead(F, ROM^, PRGBytes, j);
  { Caricamento della VROM }
  BlockRead(F, VROM^, CHRbytes, j);
  CloseFile(F);
  Result := true;
end; { LoadNES }

procedure TfrmNES.CheckKb;
begin
  Joypad1 := 0;
  { Joypad1 }
  DXInput1.Update;
  { KEY A }
  if (isButton1 in DXInput1.States) then
    Joypad1 := Joypad1 or JOY_A;
  { KEY B }
  if (isButton2 in DXInput1.States) then
    Joypad1 := Joypad1 or JOY_B;
  { KEY SELECT }
  if (isButton3 in DXInput1.States) then
    Joypad1 := Joypad1 or JOY_SELECT;
  { KEY START }
  if (isButton4 in DXInput1.States) then
    Joypad1 := Joypad1 or JOY_START;
  { KEY UP }
  if (isUp in DXInput1.States) then
    Joypad1 := Joypad1 or JOY_UP;
  { KEY DOWN }
  if (isDown in DXInput1.States) then
    Joypad1 := Joypad1 or JOY_DOWN;
  { KEY LEFT }
  if (isLeft in DXInput1.States) then
    Joypad1 := Joypad1 or JOY_LEFT;
  { KEY RIGHT }
  if (isRight in DXInput1.States) then
    Joypad1 := Joypad1 or JOY_RIGHT;
  (*
  { Joypad2 }
  DXInput2.Update;
  { KEY A }
  if (isButton1 in DXInput2.States) then
    WriteBit(Joypad2, 0, true)
  else
    WriteBit(Joypad2, 0, false);
  { KEY B }
  if (isButton2 in DXInput2.States) then
    WriteBit(Joypad2, 1, true)
  else
    WriteBit(Joypad2, 1, false);
  { KEY SELECT }
  if (isButton3 in DXInput2.States) then
    WriteBit(Joypad2, 2, true)
  else
    WriteBit(Joypad2, 2, false);
  { KEY START }
  if (isButton4 in DXInput2.States) then
    WriteBit(Joypad2, 3, true)
  else
    WriteBit(Joypad2, 3, false);
  { KEY UP }
  if (isUp in DXInput2.States) then
    WriteBit(Joypad2, 4, true)
  else
    WriteBit(Joypad2, 4, false);
  { KEY DOWN }
  if (isDown in DXInput2.States) then
    WriteBit(Joypad2, 5, true)
  else
    WriteBit(Joypad2, 5, false);
  { KEY LEFT }
  if (isLeft in DXInput2.States) then
    WriteBit(Joypad2, 6, true)
  else
    WriteBit(Joypad2, 6, false);
  { KEY RIGHT }
  if (isRight in DXInput2.States) then
    WriteBit(Joypad2, 7, true)
  else
    WriteBit(Joypad2, 7, false);
  *)
end; { CheckKb }

procedure TfrmNES.PlayLogo;
var
  Line,
  FAngle : integer;
begin
  FAngle := 0;
  repeat
    if DXDraw.CanDraw then
    begin
      DXDraw.Surface.Fill(0);
      DXImageList.Items[0].DrawRotateAlpha(DXDraw.Surface, DXDraw.Surface.Width div 2,
        DXDraw.Surface.Height div 2, DXImageList.Items[0].Width, DXImageList.Items[0].Height,
        0, 0.5, 0.5, FAngle, Trunc(Cos256(FAngle)*126+127));
      DXDraw.Flip;
    end;
    inc(FAngle, 2);
  until FAngle>130;
  { Cancello la memoria video }
  for Line := MINLINE to MAXLINE do
    FillChar(DXDIB.DIB.ScanLine[Line]^, DXDIB.DIB.WidthBytes, BLACK_COLOR);
end; { PlayLogo }

procedure TfrmNES.DrawScreen;
var
  r : double;
  t : longint;
begin
  if not DXDraw.CanDraw then exit;
  with DXDraw.Surface.Canvas do
  begin
    CopyRect(Rect(0, 0, ScreenWidth, ScreenHeight), DXDIB.DIB.Canvas,
             Rect(0, 0, //MINLINE,
                  NES_SCREEN_WIDTH - 1,
                  NES_SCREEN_HEIGHT - 1)
                  //MAXLINE - MINLINE + 1)
             );
    inc(CurrentFrame);
    if ViewFreq then
    begin
      t := GetTickCount;
     if (LastTimerTick<>0) and (t<>LastTimerTick) then
       R := (CurrentFrame * 1000) / (t - LastTimerTick);
     LastTimerTick := t;
     CurrentFrame := 0;

      Brush.Style := bsClear;
      Font.Color := clWhite;
      Font.Size := 8;
      Textout(0, 0, format('%2.2f FPS', [r]));
    end;
    Release;
  end;

  DXDraw.Flip;
end; { DrawScreen }

procedure TfrmNES.RefreshLine(Y1, Y2 : integer);
var
  iPP,
  Line : integer;
  PP: ^TByteArray;
  nametbl,
  x, y, X1, vofs,h , ntbl : integer;
  no, atr, cc, c1, c2 : byte;
  vadr, i : integer;
begin
  inc(Y2);
  if (PPU[1] and $8)=0 then
  begin
    for Line := Y1 to Y2 do
    begin
      PP := DXDIB.DIB.ScanLine[Line];
      FillChar(PP^, DXDIB.DIB.WidthBytes, BLACK_COLOR);
    end;
  end
  else
  begin
    vofs := 0;
    if (PPU[0] and $10)<>0 then
      vofs := $1000;
    y := Y1 + PPUSCROLL_Y1;
    h := 8 - (y and 7);
    inc(vofs, (y and 7));
    y := y shr 3;
    ntbl := $2000 + (PPU[0] and $3) * $400;
    Line := Y1;
    iPP := 0;
    repeat
      x := PPUSCROLL_X1 shr 3;
      if (y>=30) then
      begin
        dec(y, 30);
        ntbl := ntbl xor $800;
      end;
      nametbl := ntbl;
      for X1 := 0 to 31 do
      begin
        no := VRAM[nametbl + x + y * 32];
        atr := VRAM[nametbl + $03C0 + ( x shr 2 ) + ( y shr 2 ) * 8];
        cc := ((atr shr ((x and 2) + (y and 2) * 2)) and 3) * 4;
        vadr := no * 16 + vofs;
        for i:=0 to (h - 1) do
        begin
          PP := DXDIB.DIB.ScanLine[Line + i];
          c1 := ((Ord(VPage[vadr shr 10][vadr]) shr 1) and $55) or (Ord(VPage[(vadr + 8) shr 10][vadr + 8]) and $AA);
          c2 := (Ord(VPage[vadr shr 10][vadr]) and $55) or ((Ord(VPage[(vadr + 8) shr 10][vadr + 8]) shl 1) and $AA);
          PP[iPP + 0] := VRAM[$3F00 + ((c1 shr 6) and 3) + cc];
          PP[iPP + 1] := VRAM[$3F00 + ((c2 shr 6) and 3) + cc];
          PP[iPP + 2] := VRAM[$3F00 + ((c1 shr 4) and 3) + cc];
          PP[iPP + 3] := VRAM[$3F00 + ((c2 shr 4) and 3) + cc];
          PP[iPP + 4] := VRAM[$3F00 + ((c1 shr 2) and 3) + cc];
          PP[iPP + 5] := VRAM[$3F00 + ((c2 shr 2) and 3) + cc];
          PP[iPP + 6] := VRAM[$3F00 + (c1 and 3) + cc];
          PP[iPP + 7] := VRAM[$3F00 + (c2 and 3) + cc];

          inc(vadr);
        end;
        inc(x);
        if x=32 then
        begin
          x := 0;
          nametbl := ntbl xor $400;
        end;
        inc(iPP, 8);
      end;
      inc(Line, h);
      iPP := 0;
      vofs := vofs and not(7);
      h := Y2 - Line;
      if (h>8) then
        h := 8;
      inc(y);
    until (Line>=Y2);
  end;
  { Refresh Sprite }
  //if (PPU[1] and $10)<>0 then
    RefreshSprite(Y1, Y2);
end; { RefreshLine }

procedure TfrmNES.RefreshSprite(Y1, Y2 : integer);
type
  TSPR = record
           y, no, atr, x : byte;
         end;
var
  n, vofs : integer;
  HH : byte;
  sprite : ^TSPR;
  PP: ^TByteArray;
  iPP, C : integer;
  x, y, no, atr, CC, J, c1, c2 : byte;
  h, t, incr, vadr : integer;
begin
  vofs := 0;
  if (PPU[0] and $20)<>0 then { Sprite 16 }
    HH := 16
  else
  begin
    HH := 8;
    if (PPU[0] and $08)<>0 then { Sprite patterm address }
      vofs := $1000;
  end;

  n := $00FC;
  repeat
    sprite := @SRAM[n];
    y := sprite.y;
    if not((y>Y2) or (y<(Y1-HH))) then
    begin
      x   := sprite.x;
      no  := sprite.no;
      atr := sprite.atr;
      cc  := (atr and $3) * 4 + 16;
      if (PPU[0] and $20)<>0 then { Sprite 16 }
        vadr := ((no and $1) shl 12) + (no shr 1)* 32
      else
        vadr := no * 16 + vofs;
      C := vadr;
      iPP := x;
      incr := 1;
      if (atr and V_FLIP)<>0 then
      begin
        incr := -1;
        inc(C, 7);
        if (PPU[0] and $20)<>0 then { Sprite 16 }
          inc(C, 16);
      end;
      t := Y1 - y;
      if t>0 then
      begin
        inc(C, t*incr);
        iPP := x;
        if (t>8) then
          inc(C, incr*8);
      end
      else
        t := 0;
      h := Y2 - y + 1;
      if (h>HH) then h := HH;
      repeat
        PP  := DXDIB.DIB.ScanLine[y + t];
        if (t=8) then
          inc(C, incr*8);
        c1 := ((Ord(VPage[C shr 10][C]) shr 1) and $55) or (Ord(VPage[(C + 8) shr 10][C + 8]) and $AA);
        c2 := (Ord(VPage[C shr 10][C]) and $55) or ((Ord(VPage[(C + 8) shr 10][C + 8]) shl 1) and $AA);
        J := Ord(VPage[C shr 10][C]) or Ord(VPage[(C + 8) shr 10][C + 8]);
        if J<>0 then
          if (atr and H_FLIP)<>0 then
          begin
            if (J and $80)<>0 then
              PP[iPP + 7] := VRAM[$3F00 + ((c1 shr 6) and 3) + cc];
            if (J and $40)<>0 then
              PP[iPP + 6] := VRAM[$3F00 + ((c2 shr 6) and 3) + cc];
            if (J and $20)<>0 then
              PP[iPP + 5] := VRAM[$3F00 + ((c1 shr 4) and 3) + cc];
            if (J and $10)<>0 then
              PP[iPP + 4] := VRAM[$3F00 + ((c2 shr 4) and 3) + cc];
            if (J and $8)<>0 then
              PP[iPP + 3] := VRAM[$3F00 + ((c1 shr 2) and 3) + cc];
            if (J and $4)<>0 then
              PP[iPP + 2] := VRAM[$3F00 + ((c2 shr 2) and 3) + cc];
            if (J and $2)<>0 then
              PP[iPP + 1] := VRAM[$3F00 + (c1 and 3) + cc];
            if (J and $1)<>0 then
              PP[iPP + 0] := VRAM[$3F00 + (c2 and 3) + cc];
          end
          else
          begin
            if (J and $80)<>0 then
              PP[iPP + 0] := VRAM[$3F00 + ((c1 shr 6) and 3) + cc];
            if (J and $40)<>0 then
              PP[iPP + 1] := VRAM[$3F00 + ((c2 shr 6) and 3) + cc];
            if (J and $20)<>0 then
              PP[iPP + 2] := VRAM[$3F00 + ((c1 shr 4) and 3) + cc];
            if (J and $10)<>0 then
              PP[iPP + 3] := VRAM[$3F00 + ((c2 shr 4) and 3) + cc];
            if (J and $8)<>0 then
              PP[iPP + 4] := VRAM[$3F00 + ((c1 shr 2) and 3) + cc];
            if (J and $4)<>0 then
              PP[iPP + 5] := VRAM[$3F00 + ((c2 shr 2) and 3) + cc];
            if (J and $2)<>0 then
              PP[iPP + 6] := VRAM[$3F00 + (c1 and 3) + cc];
            if (J and $1)<>0 then
              PP[iPP + 7] := VRAM[$3F00 + (c2 and 3) + cc];
          end;
        inc(t);
        inc(C, incr);
      until t>=h;
    end;
    dec(n, $04);
  until n<$0000;
end; { RefreshSprite }

procedure TfrmNES.FormCreate(Sender: TObject);
var
  i : integer;
  NESRGBQuads : TRGBQuads;
begin
  SoundOn := false;
  CPUPeriod := 115;
  RomFilename := '';
  MRUList := TStringList.Create;
  { inizializza cpu }
  with cpu do
  begin
    A := 0;
    P := 0;
    X := 0;
    Y := 0;
    S := 0;
    PC.W  := 0;
    iPeriod  := CPUPeriod;
    ICount   := 0;
    IRequest := 0;
    AfterCLI := 0;
    IBackup  := 0;
    //void *User;
    TrapBadOps := 0;
    Trap  := 0;
    Trace := 0;
  end;
  {
  cpu := TM6502.Create(Self);
  with cpu do
  begin
    MemoryWrite      := Self.MemoryWrite;
    MemoryReadByte   := Self.MemoryReadByte;
    MemoryReadOpcode := Self.MemoryReadOpcode;
    OnLoop           := cpuLoop;
    OnStop           := TrashMachine;
  end;
  }
  { legge le risorse }
  VersionResource.FileName := Application.EXEname;
  Version := VersionResource.VersionInfo.ProductVersion;
  { Read configuration }
  ReadConfig;
  with DXInput1.Keyboard do
  begin
    FillChar(KeyAssigns, SizeOf(KeyAssigns), 0);
    AssignKey(KeyAssigns, isUp,      [BUTTON1_UP]);
    AssignKey(KeyAssigns, isDown,    [BUTTON1_DOWN]);
    AssignKey(KeyAssigns, isLeft,    [BUTTON1_LEFT]);
    AssignKey(KeyAssigns, isRight,   [BUTTON1_RIGHT]);
    AssignKey(KeyAssigns, isButton1, [BUTTON1_A]);
    AssignKey(KeyAssigns, isButton2, [BUTTON1_B]);
    AssignKey(KeyAssigns, isButton3, [BUTTON1_START]);
    AssignKey(KeyAssigns, isButton4, [BUTTON1_SELECT]);
  end;
  DXInput1.Joystick.Enabled := BUTTON1_ENABLED;
  with DXInput2.Keyboard do
  begin
    FillChar(KeyAssigns, SizeOf(KeyAssigns), 0);
    AssignKey(KeyAssigns, isUp,      [BUTTON2_UP]);
    AssignKey(KeyAssigns, isDown,    [BUTTON2_DOWN]);
    AssignKey(KeyAssigns, isLeft,    [BUTTON2_LEFT]);
    AssignKey(KeyAssigns, isRight,   [BUTTON2_RIGHT]);
    AssignKey(KeyAssigns, isButton1, [BUTTON2_A]);
    AssignKey(KeyAssigns, isButton2, [BUTTON2_B]);
    AssignKey(KeyAssigns, isButton3, [BUTTON2_START]);
    AssignKey(KeyAssigns, isButton4, [BUTTON2_SELECT]);
  end;
  DXInput2.Joystick.Enabled := BUTTON2_ENABLED;
  //InitScreen;
  { initialize }
  Randomize;
  Terminal.Lines.Add(VersionResource.VersionInfo.ProductName+ ' ' + Version);
  Terminal.Lines.Add(VersionResource.VersionInfo.FileDescription); //'The NES Emulator');
  Terminal.Lines.Add(VersionResource.VersionInfo.LegalCopyright);  //'(C) 1998 by Dott. Cosimo De Michele');
  Terminal.Lines.Add(VersionResource.VersionInfo.LegalTrademarks);
  Terminal.Lines.Add('');
  { inizializzo variabili }
  CurrentFrame := 1;
  ROM := nil;
  VROM := nil;
  { Creo la palette }
  for i:=0 to 63 do
    NESRGBQuads[i] := RGBQuad(Pal[i, 1], Pal[i, 2], Pal[i, 3]);
  { inizializzo buffer video }
  DXDIB.DIB.BitCount := 8;
  DXDIB.DIB.ColorTable := NESRGBQuads;
  DXDIB.DIB.Width    := NES_SCREEN_WIDTH;
  DXDIB.DIB.Height   := NES_SCREEN_HEIGHT;
  DXDIB.DIB.UpdatePalette;
  DXDraw.ColorTable := DXDIB.DIB.ColorTable;
  DXDraw.DefColorTable := DXDIB.DIB.ColorTable;
  DXDraw.UpdatePalette;
end;

procedure TfrmNES.FormDestroy(Sender: TObject);
begin
  //cpu.Free;
  WriteConfig;
  MRUList.free;
end;

procedure TfrmNES.mnuExitClick(Sender: TObject);
begin
  Close;
end; { mnuExitClick }

function TfrmNES.MemoryReadByte(Addr: Word): Byte;
begin
  case (Addr shr 13) of
    0 : { $0000 -- RAM }
        Result := RAM[Addr and $7FF];
    1,  { $2000 -- PPU }
    2 : { $4000 -- PSG }
        Result := IORead(Addr);
    3 : { $6000 -- WRAM }
        Result := WRAM[Addr and $1FFF];
  else
    { $8000 -- ROM }
    Result := Ord(Page[Addr shr 13][Addr]);
  end;
end; { MemoryReadByte }

procedure TfrmNES.MemoryWrite(Addr: Word; value: Byte);
begin
  case (Addr shr 13) of
    0 : { $0000 -- RAM }
        RAM[Addr and $7FF] := value;
    1,  { $2000 -- PPU }
    2 : { $4000 -- PSG }
        IOWrite(Addr, value);
    3 : { $6000 -- WRAM }
        WRAM[Addr and $1FFF] := value;
  else
    { $8000 -- ROM }
    MMCWrite(Addr, value);
  end;
end; { MemoryWrite }

function TfrmNES.MemoryReadOpcode(Addr: Word): Byte;
begin
  Result := Ord(Page[Addr shr 13][Addr]);
end; { MemoryReadOpcode }

function  TfrmNES.IORead(A : word): byte;
begin
  case A of
    { PPU }
    $2002 : begin { PPU Status }
              Result := PPU[2];
              PPU[2] := PPU[2] and not(PPU_InVBlank or PPU_SpHit);
            end;
    $2007 : begin
              if PPUADDR_FirstRead then
              begin
                PPUADDR_FirstRead := false;
                Result := NO_DATA;
                Exit;
              end;
              VAddr := VAddr and $3FFF;
              if (VAddr<$2000) then
                Result := Ord(VPage[VAddr shr 10][VAddr])
              else
                Result := VRAM[VAddr];
              if (PPU[0] and $04)<>0 then { INC32 }
                inc(VAddr, 32)
              else
                inc(VAddr);
            end;
    { PGS }
    $4000..$4015 :
            begin
              Result := PSG[A - $4000];
            end;
    { Joypad }
    $4016 : begin
              Result := ((Joypad1 shr Joypad1Read) and $01);
              Joypad1Read := (Joypad1Read + 1) and 23//$07;
            end;
     $4017 : begin
              Result := (Joypad2 shr Joypad2Read) and $01;
              Joypad2Read := (Joypad2Read + 1) and 23//$07;
            end;
    else
    begin
      Terminal.Lines.Add(format('Ignore read at address $%4x', [A]));
      Result := NO_DATA;
    end;
  end;
end; { IORead }

procedure TfrmNES.IOWrite(A : word; V : byte);
var
  len : integer;
  freq,
  ch : WORD;
begin
  case A of
    $2000 : PPU[0] := V;
    $2001 : PPU[1] := V;
    $2002 : PPU[2] := V;
    $2003 : PPU[3] := V;
    $2004 : begin
              SRAM[PPU[3]] := V;
              inc(PPU[3]);
            end;
    $2005 : begin
              if not(PPUSCROLL_Toggle) then
                PPUSCROLL_X2 := V
              else if V<=239 then
                PPUSCROLL_Y2 := V;
              PPUSCROLL_Toggle := not(PPUSCROLL_Toggle);
              PPUSCROLL_Change := true
            end;
    $2006 : begin
              VAddr := (VAddr shl 8) or V;
              PPUADDR_FirstRead := true;
            end;
    $2007 : begin
              if (PPU[2] and PPU_WriteIgnore)=0 then
              begin
                VAddr := VAddr and $3FFF;
                VRAM[VAddr] := V;
                if ((VAddr and $3000)=$2000) then { Mirroring }
                  VRAM[(VAddr xor MirrorXor) ] := V
                else if (VAddr>=$3F00) and (VAddr<=$3F1F) then
                begin
                   if (VAddr=$3F10) then
                   begin
                     VRAM[$3F00] := V;
                     VRAM[$3F04] := V;
                     VRAM[$3F08] := V;
                     VRAM[$3F0C] := V;
                     VRAM[$3F10] := V;
                     VRAM[$3F14] := V;
                     VRAM[$3F18] := V;
                     VRAM[$3F1C] := V;
                   end;
                   PPU_PalChange := true
                end;
                if (PPU[0] and $04)<>0 then { INC32 }
                  inc(VAddr, 32)
                else
                  inc(VAddr);
              end;
              PPUADDR_FirstRead := false;
            end;
    { PSG }
    $4000,
    $4004,
    $4008,
    $400C : begin
              PSG[A - $4000] := V;
            end;
    $4001 : begin
              PSG[A - $4000] := V;
              if (V and $80)<>0 then
                SndFreq[0] := 1
              else
                SndFreq[0] := 0;
            end;
    $4005 : begin
              PSG[A - $4000] := V;
              if (V and $80)<>0 then
                SndFreq[1] := 1
              else
                SndFreq[1] := 0;
            end;
    $4009 : begin
              PSG[A - $4000] := V;
              if (V and $80)<>0 then
                SndFreq[2] := 1
              else
                SndFreq[2] := 0;
            end;
    $400D : begin
              PSG[A - $4000] := V;
              if (V and $80)<>0 then
                SndFreq[3] := 1
              else
                SndFreq[3] := 0;
            end;
    $4002,
    $4003 : begin
              A := A - $4000;
              PSG[A] := V;
              SndFreqVal[0] := ((PSG[$03] and 7) shl 8) or PSG[$02];
              freq := SndFreqVal[0];
              freq := PSG_BASE div (freq + 1);
              len  := (PSG[$03] shr 3);
              if (SndOff[0]) or (len=0) then
                SetFreq(0, freq, 0)
              else
                SetFreq(0, freq, SndVolume);
              SndLen[0] := len * 16 + 8;
            end;
    $4006,
    $4007 : begin
              A := A - $4000;
              PSG[A] := V;
              SndFreqVal[1] := ((PSG[$07] and 7) shl 8) or PSG[$06];
              freq := SndFreqVal[1];
              freq := PSG_BASE div (freq + 1);
              len  := (PSG[$07] shr 3);
              if (SndOff[1]) or (len=0) then
                SetFreq(1, freq, 0)
              else
                SetFreq(1, freq, SndVolume);
              SndLen[1] := len * 16 + 8;
            end;
    $400A,
    $400B : begin
              A := A - $4000;
              PSG[A] := V;
              SndFreqVal[2] := ((PSG[$0B] and 7) shl 8) or PSG[$0A];
              freq := SndFreqVal[2];
              freq := PSG_BASE div (freq + 1);
              len  := (PSG[$0B] shr 3);
              if (SndOff[2]) or (len=0) then
                SetFreq(2, freq, 0)
              else
                SetFreq(2, freq, SndVolume);
              SndLen[2] := len + 2;
            end;
    $400E,
    $400F : begin
              A := A - $4000;
              PSG[A] := V;
              SndFreqVal[3] := ((PSG[$0F] and 7) shl 8) or PSG[$0E];
              freq := SndFreqVal[3];
              freq := PSG_BASE div (freq + 1);
              len  := (PSG[$0F] shr 3);
              if (SndOff[3]) or (len=0) then
                SetFreq(3, freq, 0)
              else
                SetFreq(3, freq, SndVolume);
              SndLen[3] := len;
            end;
    { PCM }
    $4010,
    $4011,
    $4012,
    $4013 : begin
              PSG[A - $4000] := V;
            end;
    { DMA }
    $4014 : begin
              if V<$80 then
                CopyMemory(@SRAM, @RAM[(V and 7) shl 8], 256)
              else
                CopyMemory(@SRAM, PChar(Page[V shr 5][V shl 8]), 256)
            end;
    $4015 : begin
              for ch:=0 to 3 do
                if ((V shr ch) and 1)<>0 then
                begin { Enable sound }

                end
                else
                begin { disable sound }
                  SndLen[ch] := 0;
                end;
              PSG[$15] := V;
            end;
    { Joypad }
    $4016 : begin
              if ((PSG[$16] and 1) and not(V and 1))<>0 then
              begin
                Joypad1Read := 0;
                Joypad2Read := 0;
              end;
              PSG[$16] := V;
            end;
    $4017 : begin
              PSG[$17] := V;
            end;
    else
    begin
      Terminal.Lines.Add(format('Ignore write at address $%4x', [A]));
    end;
  end;
end; { IOWrite }

function TfrmNES.cpuLoop : byte;
var
  ch,
  freq : integer;
  freqS : byte;
begin
  { gestione suono }
  (*
  dec(SndCount);
  if SndCount<=0 then
  begin
    SndCount := SndPeriod;
    for ch := 0 to 3 do
    begin
      if SndLen[ch]<>0 then
      begin
        freqS := (((PSG[ch * 4 + 1] shr 4) and $07)) * ((PSG[ch + 4 + 1] and $07));
        if (PSG[ch * 4 + 1] and $08)<>0 then
          dec(SndFreqVal[ch], freqS)
        else
          inc(SndFreqVal[ch], freqS);
        freq := PSG_BASE div (SndFreqVal[ch] + 1);
        SetFreq(ch, freq and $03FF, SndVolume);
      end;
      dec(SndLen[ch]);
      if SndLen[ch]=0 then
        PSG[$15] := PSG[$15] and not(1 shl ch);
    end;
  end;
  *)
  ScanLine := (ScanLine + 1) mod SCANLINEPERFRAME;
  Result := INT_NONE;

  if (ScanLine = MINLINE) then
  begin
    PPU[2]:= PPU[2] and not(PPU_InVBlank);
    PrevLine := MINLINE;
    HitY := SRAM[0] + 8;
  end
  else if (ScanLine = MAXLINE) then
  begin
    if UCount<>0 then
      dec(UCount)
    else
    begin
      RefreshLine(PrevLine, ScanLine);
      PrevLine := ScanLine + 1;
      UCount := 2;
      DrawScreen;
    end;
  end;
  if (ScanLine>=MINLINE) and (ScanLine<=MAXLINE) then
  begin
    if PPUSCROLL_Change then
    begin
      if UCount=0 then
        RefreshLine(PrevLine, ScanLine - 1);
      PrevLine := ScanLine;
      PPUSCROLL_X1 := PPUSCROLL_X2;
      PPUSCROLL_Y1 := PPUSCROLL_Y2;
      PPUSCROLL_Change := false;
    end;
    if ((PPU[1] and $10)<>0) and (ScanLine=HitY) then
    begin
      PPU[2] := PPU[2] or PPU_SpHit;
      if (PPU[0] and $40)<>0 then { Sprite Hit ON }
        Result := INT_NMI;
    end;
  end
  else if ScanLine=240 then
  begin
    { Check the Keyboard }
    CheckKb;
    PPU[2] := (PPU[2] and not(PPU_SpHit)) or PPU_InVBlank;
    PSG[$16] := PSG[$16] and $FD;
    if (PPU[0] and $80)<>0 then { VBlank ON }
      Result := INT_NMI;
    Application.ProcessMessages;
    if not NESisRunning then
      Result := INT_QUIT;
  end;
end; { cpuLoop }

procedure TfrmNES.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
var
  s : string;
  i, AWidth, AHeight, ABitCount: Integer;
begin
  case Key of
    VK_ESCAPE : Close;
    VK_F12    : begin
                  DXDraw.Finalize;
                  if doFullScreen in DXDraw.Options then
                  begin { from FullScreen to normal screen }
                    RestoreWindow;
                    Menu := MainMenu;
                    StatusBar.visible := true;
                    BorderStyle := bsSingle;
                    DXDraw.Cursor := crDefault;
                    DXDraw.Options := DXDraw.Options - [doFullScreen];
                  end
                  else
                  begin { from normal screen to FullScreen }
                    StoreWindow;
                    Menu := nil;
                    StatusBar.visible := false;
                    BorderStyle := bsNone;
                    DXDraw.Cursor := crNone;
                    if SCREEN_MODE<>'' then
                    begin
                      s := SCREEN_MODE;
                      i := Pos('x', s);
                      AWidth := StrToInt(Copy(s, 1, i-1));
                      s := Copy(s, i+1, Length(s));
                      i := Pos('x', s);
                      AHeight := StrToInt(Copy(s, 1, i-1));
                      s := Copy(s, i+1, Length(s));
                      ABitCount := StrToInt(s);
                      DXDraw.Display.Width := AWidth;
                      DXDraw.Display.Height := AHeight;
                      DXDraw.Display.BitCount := ABitCount;
                    end;
                    DXDraw.Options := DXDraw.Options + [doFullScreen];
                  end;
                  DXDraw.Initialize;
                end;
    VK_F2     : mnuLoadROMClick(mnuLoadROM);
    VK_F3     : mnuFreeROMClick(mnuFreeROM);
  end;
end;

procedure TfrmNES.mnuDebugClick(Sender: TObject);
begin
  {cpu.Pause;
  frmDebug.ShowModal;
  cpu.Run;}
end; { mnuDebugClick }

procedure TfrmNES.mnuPauseClick(Sender: TObject);
begin
  //cpu.Pause;
  if NESisRunning then
    NESisRunning := false
  else
  begin
    NESisRunning := true;
    Run6502(cpu);
  end;
end;

procedure TfrmNES.mnuResetClick(Sender: TObject);
begin
  //cpu.Pause;
  NESisRunning := false;
  ResetNes;
  NESisRunning := true;
  Run6502(cpu);
end;

procedure TfrmNES.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  mnuFreeROM.Click;
end;

procedure TfrmNES.mnuRomClick(Sender: TObject);
var
  frmROMProperty: TfrmROMProperty;
begin
 frmROMProperty := TfrmROMProperty.Create(Self);
 with frmROMProperty do
 begin
   edFilename.Text := RomFilename;
   edHeader.Text   := RomHeader.RomType;
   edProgPages.Text:= IntToStr(Ord(RomHeader.PRGROM));
   lblROM.Caption  := format('[%d Kb]', [Ord(RomHeader.PRGROM) * 16]);
   edCharPages.Text:= IntToStr(Ord(RomHeader.CHRRAM));
   lblVROM.Caption := format('[%d Kb]', [Ord(RomHeader.CHRRAM) * 8]);
   edMapping.Text  := '#' + IntToStr(MMCType and 15);
   cbVertMirror.Checked := (Mirroring = 0);
   cbBattery.Checked    := (ROMType and ROM_WRAM)<>0;
   cbTrainer.Checked    := (ROMType and ROM_TRAINER)<>0;
   cbFourScreen.Checked := (ROMType and ROM_FOUR)<>0;
   edExpansion.Text := ROMHeader.ROMctrl2 + RomHeader.Expansion;
   ShowModal;
 end;
end;

procedure TfrmNES.mnuAboutClick(Sender: TObject);
var
  frmAbout: TfrmAbout;
begin
  frmAbout := TfrmAbout.Create(Application);
  frmAbout.ShowModal;
  frmAbout.free;
end;

procedure TfrmNES.mnuFreeROMClick(Sender: TObject);
begin
  mnuRom.enabled := false;
  mnuFreeROM.enabled := false;
  mnuSaveSnapShot.enabled := true;
  //cpu.Stop;
  NESisRunning := false;
  TrashMachine;
end;

procedure TfrmNES.mnuInputClick(Sender: TObject);
var
  frmSetting: TfrmSetting;
begin
  frmSetting := TfrmSetting.Create(Self);
  frmSetting.ShowModal;
  if frmSetting.ModalResult = mrOK then
  begin
    { Joypad 1 }
    with DXInput1.Keyboard do
    begin
      FillChar(KeyAssigns, SizeOf(KeyAssigns), 0);
      AssignKey(KeyAssigns, isUp,      [BUTTON1_UP]);
      AssignKey(KeyAssigns, isDown,    [BUTTON1_DOWN]);
      AssignKey(KeyAssigns, isLeft,    [BUTTON1_LEFT]);
      AssignKey(KeyAssigns, isRight,   [BUTTON1_RIGHT]);
      AssignKey(KeyAssigns, isButton1, [BUTTON1_A]);
      AssignKey(KeyAssigns, isButton2, [BUTTON1_B]);
      AssignKey(KeyAssigns, isButton3, [BUTTON1_START]);
      AssignKey(KeyAssigns, isButton4, [BUTTON1_SELECT]);
    end;
    DXInput1.Joystick.Enabled := BUTTON1_ENABLED;
    { Joypad 2 }
    with DXInput2.Keyboard do
    begin
      FillChar(KeyAssigns, SizeOf(KeyAssigns), 0);
      AssignKey(KeyAssigns, isUp,      [BUTTON2_UP]);
      AssignKey(KeyAssigns, isDown,    [BUTTON2_DOWN]);
      AssignKey(KeyAssigns, isLeft,    [BUTTON2_LEFT]);
      AssignKey(KeyAssigns, isRight,   [BUTTON2_RIGHT]);
      AssignKey(KeyAssigns, isButton1, [BUTTON2_A]);
      AssignKey(KeyAssigns, isButton2, [BUTTON2_B]);
      AssignKey(KeyAssigns, isButton3, [BUTTON2_START]);
      AssignKey(KeyAssigns, isButton4, [BUTTON2_SELECT]);
    end;
    DXInput2.Joystick.Enabled := BUTTON2_ENABLED;
    ViewFreq := frmSetting.ckViewFreq.Checked;
    //InitScreen;
  end;
  frmSetting.free;
end;

procedure TfrmNES.ReadConfig;
var
  Reg : TRegistry;
begin
  Reg := TRegistry.Create;
  if Reg.OpenKey(RegKey, true) then
  begin
    { NES }
    if Reg.ValueExists('FrameRate') then
      FrameRate := Reg.ReadInteger('FrameRate')
    else
      FrameRate := 3;
    if Reg.ValueExists('ViewFrequency') then
      ViewFreq := Reg.ReadBool('ViewFrequency')
    else
      ViewFreq := false;
    { Joypad 1 }
    if Reg.ValueExists('BUTTON1 ENABLED') then
      BUTTON1_ENABLED := Reg.ReadBool('BUTTON1 ENABLED')
    else
      BUTTON1_ENABLED := false;
    if Reg.ValueExists('BUTTON1 UP') then
      BUTTON1_UP := Reg.ReadInteger('BUTTON1 UP')
    else
      BUTTON1_UP := VK_UP;
    if Reg.ValueExists('BUTTON1 DOWN') then
      BUTTON1_DOWN := Reg.ReadInteger('BUTTON1 DOWN')
    else
      BUTTON1_DOWN := VK_DOWN;
    if Reg.ValueExists('BUTTON1 LEFT') then
      BUTTON1_LEFT := Reg.ReadInteger('BUTTON1 LEFT')
    else
      BUTTON1_LEFT := VK_LEFT;
    if Reg.ValueExists('BUTTON1 RIGHT') then
      BUTTON1_RIGHT := Reg.ReadInteger('BUTTON1 RIGHT')
    else
      BUTTON1_RIGHT := VK_RIGHT;
    if Reg.ValueExists('BUTTON1 A') then
      BUTTON1_A := Reg.ReadInteger('BUTTON1 A')
    else
      BUTTON1_A := VK_PRIOR;
    if Reg.ValueExists('BUTTON1 B') then
      BUTTON1_B := Reg.ReadInteger('BUTTON1 B')
    else
      BUTTON1_B := VK_NEXT;
    if Reg.ValueExists('BUTTON1 START') then
      BUTTON1_START := Reg.ReadInteger('BUTTON1 START')
    else
      BUTTON1_START := VK_RETURN;
    if Reg.ValueExists('BUTTON1 SELECT') then
      BUTTON1_SELECT := Reg.ReadInteger('BUTTON1 SELECT')
    else
      BUTTON1_SELECT := VK_SPACE;
    { Joypad 2 }
    if Reg.ValueExists('BUTTON2 ENABLED') then
      BUTTON2_ENABLED := Reg.ReadBool('BUTTON2 ENABLED')
    else
      BUTTON2_ENABLED := false;
    if Reg.ValueExists('BUTTON2 UP') then
      BUTTON2_UP := Reg.ReadInteger('BUTTON2 UP')
    else
      BUTTON2_UP := VK_UP;
    if Reg.ValueExists('BUTTON2 DOWN') then
      BUTTON2_DOWN := Reg.ReadInteger('BUTTON2 DOWN')
    else
      BUTTON2_DOWN := VK_DOWN;
    if Reg.ValueExists('BUTTON2 LEFT') then
      BUTTON2_LEFT := Reg.ReadInteger('BUTTON2 LEFT')
    else
      BUTTON2_LEFT := VK_LEFT;
    if Reg.ValueExists('BUTTON2 RIGHT') then
      BUTTON2_RIGHT := Reg.ReadInteger('BUTTON2 RIGHT')
    else
      BUTTON2_RIGHT := VK_RIGHT;
    if Reg.ValueExists('BUTTON2 A') then
      BUTTON2_A := Reg.ReadInteger('BUTTON2 A')
    else
      BUTTON2_A := VK_PRIOR;
    if Reg.ValueExists('BUTTON2 B') then
      BUTTON2_B := Reg.ReadInteger('BUTTON2 B')
    else
      BUTTON2_B := VK_NEXT;
    if Reg.ValueExists('BUTTON2 START') then
      BUTTON2_START := Reg.ReadInteger('BUTTON2 START')
    else
      BUTTON2_START := VK_RETURN;
    if Reg.ValueExists('BUTTON2 SELECT') then
      BUTTON2_SELECT := Reg.ReadInteger('BUTTON2 SELECT')
    else
      BUTTON2_SELECT := VK_SPACE;
    { Screen }
    if Reg.ValueExists('WIN') then
      WIN_MODE := Reg.ReadInteger('WIN')
    else
      WIN_MODE := 0; // Normal Screen
    if Reg.ValueExists('SCREEN') then
      SCREEN_MODE := Reg.ReadString('SCREEN')
    else
      SCREEN_MODE := ''; // Default Screen
    { Most Recent Used }
    MRUList.Add(Reg.ReadString('MRU1'));
    MRUList.Add(Reg.ReadString('MRU2'));
    MRUList.Add(Reg.ReadString('MRU3'));
    MRUList.Add(Reg.ReadString('MRU4'));
    MRUList.Add(Reg.ReadString('MRU5'));
  end;
  Reg.free;
end; { ReadConfig }

procedure TfrmNES.WriteConfig;
var
  Reg : TRegistry;
  i : integer;
begin
  Reg := TRegistry.Create;
  if Reg.OpenKey('Software\Karakiri\WinNES', true) then
  begin
    { NES }
    Reg.WriteInteger('FrameRate', FrameRate);
    Reg.WriteBool('ViewFrequency', ViewFreq);
    { Joypad 1 }
    Reg.WriteInteger('BUTTON1 UP', BUTTON1_UP);
    Reg.WriteInteger('BUTTON1 DOWN', BUTTON1_DOWN);
    Reg.WriteInteger('BUTTON1 LEFT', BUTTON1_LEFT);
    Reg.WriteInteger('BUTTON1 RIGHT', BUTTON1_RIGHT);
    Reg.WriteInteger('BUTTON1 A', BUTTON1_A);
    Reg.WriteInteger('BUTTON1 B', BUTTON1_B);
    Reg.WriteInteger('BUTTON1 START', BUTTON1_START);
    Reg.WriteInteger('BUTTON1 SELECT', BUTTON1_SELECT);
    Reg.WriteBool('BUTTON1 ENABLED', BUTTON1_ENABLED);
    { Joypad 2 }
    Reg.WriteInteger('BUTTON2 UP', BUTTON2_UP);
    Reg.WriteInteger('BUTTON2 DOWN', BUTTON2_DOWN);
    Reg.WriteInteger('BUTTON2 LEFT', BUTTON2_LEFT);
    Reg.WriteInteger('BUTTON2 RIGHT', BUTTON2_RIGHT);
    Reg.WriteInteger('BUTTON2 A', BUTTON2_A);
    Reg.WriteInteger('BUTTON2 B', BUTTON2_B);
    Reg.WriteInteger('BUTTON2 START', BUTTON2_START);
    Reg.WriteInteger('BUTTON2 SELECT', BUTTON2_SELECT);
    Reg.WriteBool('BUTTON2 ENABLED', BUTTON2_ENABLED);    
    { Screen }
    Reg.WriteString('SCREEN', SCREEN_MODE);
    Reg.WriteInteger('WIN', WIN_MODE);
    { Most recently used }
    for i:=0 to (MRUCount-1) do
      Reg.WriteString(format('MRU%d', [i + 1]), MRUList.Strings[i]);
  end;
  Reg.free;
end; { WriteConfig }

procedure TfrmNES.InitScreen;
var
  s : string;
  i,
  AWidth,
  AHeight,
  ABitCount : integer;
begin
  ScreenWidth  := NES_SCREEN_WIDTH;
  ScreenHeight := NES_SCREEN_HEIGHT;
  StoreWindow;
  case WIN_MODE of
    1 : begin
          ScreenWidth  := ScreenWidth * 2;
          ScreenHeight := ScreenHeight * 2;
          if ClientWidth < ScreenWidth then
            ClientWidth  := ScreenWidth;
          if (ClientHeight - StatusBar.Height) < ScreenHeight then
            ClientHeight := ScreenHeight + StatusBar.Height;
        end;
    2 : begin
          Menu := nil;
          StatusBar.visible := false;
          BorderStyle := bsNone;
          DXDraw.Cursor := crNone;
          if SCREEN_MODE<>'' then
          begin
            s := SCREEN_MODE;
            i := Pos('x', s);
            AWidth := StrToInt(Copy(s, 1, i-1));
            s := Copy(s, i+1, Length(s));
            i := Pos('x', s);
            AHeight := StrToInt(Copy(s, 1, i-1));
            s := Copy(s, i+1, Length(s));
            ABitCount := StrToInt(s);
            DXDraw.Display.Width := AWidth;
            DXDraw.Display.Height := AHeight;
            DXDraw.Display.BitCount := ABitCount;
          end;
          DXDraw.Options := DXDraw.Options + [doFullScreen];
        end;
  end;
  { inizializzo video }
  DXDraw.SurfaceWidth  := ScreenWidth;
  DXDraw.SurfaceHeight := ScreenHeight;
  Terminal.visible := false;
  DXDraw.visible := true;
  DXDraw.Initialize;
  { Play Logo }
  PlayLogo;
end; { InitScreen }

procedure TfrmNES.TrashScreen;
begin
  DXDraw.Finalize;
  Terminal.visible := true;
  DXDraw.visible := false;
  Menu := MainMenu;
  StatusBar.visible := true;
  BorderStyle := bsSingle;
  DXDraw.Cursor := crDefault;
  DXDraw.Options := DXDraw.Options - [doFullScreen];
  RestoreWindow;
end; { TrashScreen }

procedure TfrmNES.mnuSavesnapshotClick(Sender: TObject);
begin
  if SaveDialog.Execute then
  begin
    DXDIB.DIB.SaveToFile(SaveDialog.Filename);
    Terminal.Lines.Add('Snapshot saved as "' + ExtractFileName(SaveDialog.Filename) + '".');
  end;
end;

procedure TfrmNES.MRUClick(Sender: TObject);
var
  index : integer;
  ItemClicked : TMenuItem;
begin
  ItemClicked := TMenuItem(Sender);
  index := ItemClicked.MenuIndex - MRU1.MenuIndex;
  RunMachine(MRUList.Strings[index]);
end;

procedure TfrmNES.mnuFileClick(Sender: TObject);
var
  i, index : integer;
begin
  index := MRU1.MenuIndex;
  for i:=0 to (MRUCount - 1) do
  begin
    mnuFile.Items[index + i].visible := (MRUList.Strings[i]<>'');
    mnuFile.Items[index + i].Caption := MRUList.Strings[i];
  end;
  MRUSeparator.visible := mnuFile.Items[index].visible;
end;

procedure TfrmNES.mnuCPUClick(Sender: TObject);
begin
  mnuReset.Enabled := RomFilename<>'';
  mnuPause.Enabled := RomFilename<>'';
  if NESisRunning then
    mnuPause.Caption := '&Pause'
  else
    mnuPause.Caption := '&Run';
end; { mnuCPUClick }

procedure TfrmNES.mnuFullScreenClick(Sender: TObject);
var
  s : string;
  i, AWidth, AHeight, ABitCount: Integer;
begin
  DXDraw.Finalize;
  if doFullScreen in DXDraw.Options then
  begin { from FullScreen to normal screen }
    RestoreWindow;
    Menu := MainMenu;
    StatusBar.visible := true;
    BorderStyle := bsSingle;
    DXDraw.Cursor := crDefault;
    DXDraw.Options := DXDraw.Options - [doFullScreen];
  end
  else
  begin { from normal screen to FullScreen }
    StoreWindow;
    Menu := nil;
    StatusBar.visible := false;
    BorderStyle := bsNone;
    DXDraw.Cursor := crNone;
    if SCREEN_MODE<>'' then
    begin
      s := SCREEN_MODE;
      i := Pos('x', s);
      AWidth := StrToInt(Copy(s, 1, i-1));
      s := Copy(s, i+1, Length(s));
      i := Pos('x', s);
      AHeight := StrToInt(Copy(s, 1, i-1));
      s := Copy(s, i+1, Length(s));
      ABitCount := StrToInt(s);
      DXDraw.Display.Width := AWidth;
      DXDraw.Display.Height := AHeight;
      DXDraw.Display.BitCount := ABitCount;
    end;
    DXDraw.Options := DXDraw.Options + [doFullScreen];
  end;
  InitScreen;
  DXDraw.Initialize;
end; { mnuFullScreenClick }

end.
